from baangt import hook_spec
from baangt.base import GlobalConstants as GC

# -------------------------------------------------------------------

# correspond to baangt/base/TestRun

# -------------------------------------------------------------------


class baangtHookSpec(object):

    @hook_spec
    def testRun_init(self, testRunName, globalSettingsFileNameAndPath):
        pass


    @hook_spec
    def testRun_tearDown(self, testRunObject):
        pass


    @hook_spec
    def testRun_getSuccessAndError(self, testRunObject):
        pass


    @hook_spec
    def testRun_getAllTestRunAttributes(self, testRunObject):
        pass


    @hook_spec
    def testRun_getBrowser(self, testRunObject, browserInstance=1, browserName=None, browserAttributes=None):
        pass


    @hook_spec
    def testRun__getBrowserInstance(self, testRunObject, browserInstance):
        pass


    @hook_spec
    def testRun_getAPI(self, testRunObject):
        pass


    @hook_spec
    def testRun_setResult(self, testRunObject, recordNumber, dataRecordResult):
        pass


    @hook_spec
    def testRun_executeTestRun(self, testRunObject):
        pass


    @hook_spec
    def testRun_executeDictSequenceOfClasses(self, testRunObject, dictSequenceOfClasses, counterName, **kwargs):
        pass


    @hook_spec
    def testRun__initTestRun(self, testRunObject):
        pass


    @hook_spec
    def testRun_loadJSONGlobals(self, testRunObject):
        pass


    @hook_spec
    def testRun__loadJSONTestRunDefinitions(self, testRunObject):
        pass


    @hook_spec
    def testRun__loadExcelTestRunDefinitions(self, testRunObject):
        pass


    @hook_spec
    def testRun___dynamicImportClasses(self, fullQualifiedImportName):
        pass


    @hook_spec
    def testRun__sanitizeTestRunNameAndFileName(self, TestRunNameInput):
        pass

    # -------------------------------------------------------------------

    # correspond to baangt/base/Timing

    # -------------------------------------------------------------------

    @hook_spec
    def timing_init(self):
        pass

    @hook_spec
    def timing_takeTime(self, timingObject, timingName, forceNew=False):
        pass

    @hook_spec
    def timing_addAttribute(self, timingObject, attribute, value, timingSection=None):
        pass

    @hook_spec
    def timing_takeTimeSumOutput(self, timingObject):
        pass

    @hook_spec
    def timing_returnTime(self, timingObject):
        pass

    @hook_spec
    def timing_returnTimeSegment(self, timingObject, segment):
        pass

    @hook_spec
    def timing_resetTime(self, timingObject):
        pass

    @hook_spec
    def timing___format_time(self, startAndEndTimeAsDict):
        pass

    # -------------------------------------------------------------------

    # correspond to baangt/base/ExportResults

    # -------------------------------------------------------------------

    # ExportResults

    # -------------------------------------------------------------------
    @hook_spec
    def exportResults_init(self,  **kwargs):
        pass


    @hook_spec
    def exportResults_exportToDataBase(self, exportResultsObject):
        pass


    @hook_spec
    def exportResults_exportResult(self, exportResultsObject, **kwargs):
        pass


    @hook_spec
    def exportResults_makeSummary(self, exportResultsObject):
        pass


    @hook_spec
    def exportResults___writeSummaryCell(self, exportResultsObject, lineHeader, lineText, row=None, format=None):
        pass


    @hook_spec
    def exportResults___getOutputFileName(self, exportResultsObject):
        pass


    @hook_spec
    def exportResults___setHeaderDetailSheet(self, exportResultsObject):
        pass


    @hook_spec
    def exportResults__exportData(self, exportResultsObject):
        pass


    @hook_spec
    def exportResults___writeCell(self, exportResultsObject, line, cellNumber, testRecordDict, fieldName, strip=False):
        pass


    @hook_spec
    def exportResults_closeExcel(self, exportResultsObject, line, cellNumber, testRecordDict, fieldName, strip=False):
        pass

    # -------------------------------------------------------------------

    # correspond to baangt/base/ExportResults

    # -------------------------------------------------------------------

    # ExcelSheetHelperFunctions

    # -------------------------------------------------------------------
    @hook_spec
    def excelSheetHelperFunctions_init(self):
        pass


    @hook_spec
    def excelSheetHelperFunctions_set_column_autowidth(self, excelSheetHelperFunctionsObject, worksheet, column):
        pass


    @hook_spec
    def excelSheetHelperFunctions_get_column_width(self, excelSheetHelperFunctionsObject, worksheet, column):
        pass

    # -------------------------------------------------------------------

    # correspond to baangt/base/ExportResults

    # -------------------------------------------------------------------

    # ExportTiming

    # -------------------------------------------------------------------


    @hook_spec
    def exportTiming_init(self, testdataRecords, sheet):
        pass


    @hook_spec
    def exportTiming_writeHeader(self, exportTimingObject):
        pass


    @hook_spec
    def exportTiming_writeLines(self, exportTimingObject):
        pass


    @hook_spec
    def exportTiming_shortenTimingValue(self, exportTimingObject, timingValue):
        pass


    @hook_spec
    def exportTiming_writeCell(self, exportTimingObject, row, col, content, format=None):
        pass


    @hook_spec
    def exportTiming_findAllTimingSections(self, exportTimingObject):
        pass


    @hook_spec
    def exportTiming_interpretTimeLog(self, exportTimingObject, lTimeLog):
        pass


    # -------------------------------------------------------------------

    # correspond to baangt/base/BrowserDriver

    # -------------------------------------------------------------------

    @hook_spec
    def browserDriver_init(self, timing=None, screenshotPath=None):
        pass


    @hook_spec
    def browserDriver_createNewBrowser(self, browserDriverObject, browserName=GC.BROWSER_FIREFOX, desiredCapabilities=None, **kwargs):
        pass


    @hook_spec
    def browserDriver___findBrowserDriverPaths(self, browserDriverObject, filename):
        pass


    @hook_spec
    def browserDriver_slowExecutionToggle(self, browserDriverObject, newSlowExecutionWaitTimeInSeconds = None):
        pass


    @hook_spec
    def browserDriver___createBrowserOptions(self, browserDriverObject, browserName, desiredCapabilities):
        pass


    @hook_spec
    def browserDriver_closeBrowser(self, browserDriverObject):
        pass


    @hook_spec
    def browserDriver__log(self, browserDriverObject, logType, logText, **kwargs):
        pass


    @hook_spec
    def browserDriver_takeScreenshot(self, browserDriverObject, screenShotPath=None):
        pass


    @hook_spec
    def browserDriver_handleIframe(self, browserDriverObject, iframe=None):
        pass

    @hook_spec
    def browserDriver_handleWindow(self, browserDriverObject, windowNumber=None, function=None):
        pass


    @hook_spec
    def browserDriver_findByAndWaitForValue(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None, iframe=None, timeout=20,
                                  optional=False):
        pass


    @hook_spec
    def browserDriver_findByAndSetText(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None, value=None, iframe=None,
                             timeout=60, optional=False):
        pass


    @hook_spec
    def browserDriver_findByAndSetTextIf(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None, value=None, iframe=None,
                               timeout=60):
        pass


    @hook_spec
    def browserDriver_findByAndSetTextValidated(self, browserDriverObject,id = None,
                           css = None,
                           xpath = None,
                           class_name = None,
                           value = None,
                           iframe = None,
                           timeout = 60,
                           retries = 5):
        pass


    @hook_spec
    def browserDriver_submit(self, browserDriverObject):
        pass


    @hook_spec
    def browserDriver_findByAndClick(self, browserDriverObject, id = None, css=None, xpath=None, class_name=None, iframe=None, timeout=20, optional=False):
        pass


    @hook_spec
    def browserDriver_findByAndClickIf(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None, iframe=None, timeout=60,
                             value=None, optional=False):
        pass


    @hook_spec
    def browserDriver_findByAndForceText(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None, value=None,
                               iframe=None, timeout=60, optional=False):
        pass


    @hook_spec
    def browserDriver_findBy(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None, iframe=None, timeout=60, loggingOn=True,
                   optional=False):
        pass


    @hook_spec
    def browserDriver_getURL(self, browserDriverObject):
        pass


    @hook_spec
    def browserDriver___tryAndRetry(self, browserDriverObject, id=None, css=None, xpath=None, class_name=None, timeout=20):
        pass


    @hook_spec
    def browserDriver_findWaitNotVisible(self, browserDriverObject, xpath=None, id=None, timeout = 90):
        pass


    @hook_spec
    def browserDriver_sleep(self, browserDriverObject, sleepTimeinSeconds):
        pass


    @hook_spec
    def browserDriver___doSomething(self, browserDriverObject, command, value=None, timeout=20, xpath=None, optional=False):
        pass


    @hook_spec
    def browserDriver_goToUrl(self, browserDriverObject, url):
        pass


    @hook_spec
    def browserDriver_goBack(self, browserDriverObject):
        pass


    @hook_spec
    def browserDriver_javaScript(self, browserDriverObject, jsText):
        pass





from baangt.base import GlobalConstants as GC
from baangt.base.Timing.Timing import Timing
from baangt.TestSteps.Exceptions import *


class TestCaseMaster:
    def __init__(self, **kwargs):
        self.name = None
        self.description = None
        self.testSteps = {}
        self.apiInstance = None
        self.numberOfParallelRuns = None
        self.testRunInstance = kwargs.get(GC.KWARGS_TESTRUNINSTANCE)
        self.testRunUtils = self.testRunInstance.testRunUtils
        self.testSequence = self.testRunUtils.getSequenceByNumber(testRunName=self.testRunInstance.testRunName,
                                                                  sequence=kwargs.get(GC.STRUCTURE_TESTCASESEQUENCE))
        self.testCaseSettings = self.testRunUtils.getTestCaseByNumber(self.testSequence,
                                                                      kwargs.get(GC.STRUCTURE_TESTCASE))
        self.testSteps = self.testCaseSettings[2][GC.STRUCTURE_TESTSTEP]
        self.testCaseType = self.testCaseSettings[1][GC.KWARGS_TESTCASETYPE]
        self.kwargs = kwargs
        self.timing : Timing = self.kwargs.get(GC.KWARGS_TIMING)
        self.timingName = self.timing.takeTime(self.__class__.__name__, forceNew=True)
        if self.testCaseType == GC.KWARGS_BROWSER:
            logger.info(f"Settings for this TestCase: {self.testCaseSettings}. "
                        f"Starting or using Browser {self.kwargs.get(GC.KWARGS_BROWSER)}")
            if self.kwargs.get(GC.KWARGS_BROWSER):
                self.browser = self.kwargs[GC.KWARGS_BROWSER] # Browser already set - most probably from parallel run
            else:
                self.browserType = self.testCaseSettings[1][GC.KWARGS_BROWSER].upper()
                self.browserSettings = self.testCaseSettings[1][GC.BROWSER_ATTRIBUTES]
                self.mobileType = self.testCaseSettings[1].get(GC.KWARGS_MOBILE)
                self.mobileApp = self.testCaseSettings[1].get(GC.KWARGS_MOBILE_APP)
                self.mobile_desired_app = {}
                self.mobile_app_setting = {}
                if self.mobileType:
                    self.mobile_desired_app[GC.MOBILE_PLATFORM_NAME] = self.testCaseSettings[1][GC.MOBILE_PLATFORM_NAME]
                    self.mobile_desired_app[GC.MOBILE_DEVICE_NAME] = self.testCaseSettings[1][GC.MOBILE_DEVICE_NAME]
                    self.mobile_desired_app[GC.MOBILE_PLATFORM_VERSION] = self.testCaseSettings[1][GC.MOBILE_PLATFORM_VERSION]
                    self.mobile_app_setting[GC.MOBILE_APP_URL]= self.testCaseSettings[1][GC.MOBILE_APP_URL]
                    self.mobile_app_setting[GC.MOBILE_APP_PACKAGE] = self.testCaseSettings[1][GC.MOBILE_APP_PACKAGE]
                    self.mobile_app_setting[GC.MOBILE_APP_ACTIVITY] = self.testCaseSettings[1][GC.MOBILE_APP_ACTIVITY]
                self.browser = self.testRunInstance.getBrowser(browserName=self.browserType, browserAttributes=self.browserSettings,
                                                               mobileType=self.mobileType, mobileApp= self.mobileApp,
                                                               desired_app = self.mobile_desired_app, mobile_app_setting = self.mobile_app_setting)
                self.kwargs[GC.KWARGS_BROWSER] = self.browser
        elif self.testCaseType == GC.KWARGS_API_SESSION:
            # FIXME: For now we're using session_number always = 1. We need to be able to run e.g. 10 sessions with
            # FIXME: Parallel API-Test.
            self.apiInstance = self.testRunInstance.getAPI()
            self.kwargs[GC.KWARGS_API_SESSION] = self.apiInstance

            # For API-Tests we assume status = Passed and actively set it to error if not.
            self.kwargs[GC.KWARGS_DATA][GC.TESTCASESTATUS] = GC.TESTCASESTATUS_SUCCESS
        self.execute()
        self.tearDown()

    def execute(self):
        try:
            self.testRunInstance.executeDictSequenceOfClasses(self.testSteps, GC.STRUCTURE_TESTSTEP, **self.kwargs)
        except baangtTestStepException as e:
            logger.info(f"Testcase {self.kwargs.get(GC.STRUCTURE_TESTSTEP,'')} failed")
            self.kwargs[GC.KWARGS_DATA][GC.TESTCASEERRORLOG] += '\n' + "Exception-Text: " + str(e)
            self.kwargs[GC.KWARGS_DATA][GC.TESTCASESTATUS] = GC.TESTCASESTATUS_ERROR
        finally:
            self._finalizeTestCase()

    def _finalizeTestCase(self):
        tcData = self.kwargs[GC.KWARGS_DATA]
        tcData[GC.TIMING_DURATION] = self.timing.takeTime(self.timingName)   # Write the End-Record for this Testcase

        tcData[GC.TIMELOG] = self.timing.returnTime()

        self.timing.resetTime()

    def _checkAndSetTestcaseStatusIfFailExpected(self):
        """
        If this Testcase is supposed to fail and failed, he's actually successful.
        If this Testcase is supposed to fail and doesn't, he's actually failed.
        @return: Directly sets the Testcasestatus accordingly.
        """
        tcData = self.kwargs[GC.KWARGS_DATA]

        if tcData.get(GC.TESTCASE_EXPECTED_ERROR_FIELD) == 'X':
            if tcData[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR:
                tcData[GC.TESTCASESTATUS] = GC.TESTCASESTATUS_SUCCESS
            elif tcData[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_SUCCESS:
                tcData[GC.TESTCASESTATUS] = GC.TESTCASESTATUS_ERROR

    def tearDown(self):
        if self.kwargs[GC.KWARGS_DATA][GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR:
            # Try taking a Screenshot
            if self.testCaseType == GC.KWARGS_BROWSER:
                self.kwargs[GC.KWARGS_DATA][GC.SCREENSHOTS] = self.kwargs[GC.KWARGS_DATA][GC.SCREENSHOTS] + '\n' +\
                                                              self.browser.takeScreenshot()

        if not self.kwargs[GC.KWARGS_DATA][GC.TESTCASESTATUS]:
            self.kwargs[GC.KWARGS_DATA][GC.TESTCASESTATUS] = GC.TESTCASESTATUS_SUCCESS

        self._checkAndSetTestcaseStatusIfFailExpected()




import baangt.base.GlobalConstants as GC
import logging

logger = logging.getLogger("pyC")

class TestRunUtils():
    def __init__(self):
        self.testRunAttributes = {}

    def setCompleteTestRunAttributes(self, testRunName:str, testRunAttributes: dict):
        self.testRunAttributes[testRunName] = testRunAttributes

    def getCompleteTestRunAttributes(self, testRunName):
        logger.info('get into getCompleteTestRunAttributes, testRunName is {}'.format(testRunName))
        return self.testRunAttributes[testRunName][GC.KWARGS_TESTRUNATTRIBUTES]

    def getSequenceByNumber(self, sequence, testRunName):
        return self.testRunAttributes[testRunName][GC.KWARGS_TESTRUNATTRIBUTES][GC.STRUCTURE_TESTCASESEQUENCE].get(sequence)

    def getTestCaseByNumber(self, sequence, testcaseNumber):
        return sequence[1][GC.STRUCTURE_TESTCASE][testcaseNumber]

    def getTestStepByNumber(self, testCase, testStepNumber):
        return testCase[2][GC.STRUCTURE_TESTSTEP].get(testStepNumber)

    def replaceGlobals(self, globals):
        """
        Will go through all testcase-Settings and replace values with values from global settings, if matched
        """
        for key, value in globals.items():
            if not "TC." in key[0:3] or len(value) == 0:
                continue
            self.testRunAttributes = TestRunUtils._recursive_replace(self.testRunAttributes, key.replace("TC.",""), value)

    @staticmethod
    def _recursive_replace(dictToBeReplaced, lKey, lValue):
        if isinstance(dictToBeReplaced, list):
            for entry in dictToBeReplaced:
                TestRunUtils._recursive_replace(entry, lKey, lValue)
        elif isinstance(dictToBeReplaced, dict):
            if lKey in dictToBeReplaced:
                logger.info(f"Due to Globals replaced value {dictToBeReplaced[lKey]} of {lKey} with value {lValue}")
                dictToBeReplaced[lKey] = lValue
            for k,v in dictToBeReplaced.items():
                TestRunUtils._recursive_replace(v, lKey, lValue)
        else:
            pass
        return dictToBeReplaced



from sqlalchemy import Column, String, Integer, DateTime
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base

DATABASE_URL = 'testrun.db'

engine = create_engine(f'sqlite:///{DATABASE_URL}')
base = declarative_base()

# declare models
class TestrunLog(base):
	__tablename__ = "testrunLog"
	id = Column(Integer, primary_key=True)
	testrunName = Column(String, nullable=False)
	logfileName = Column(String, nullable=False)
	startTime = Column(DateTime, nullable=False)
	endTime = Column(DateTime, nullable=False)
	globalVars = Column(String, nullable=True)
	dataFile = Column(String, nullable=True)
	statusOk = Column(Integer, nullable=False)
	statusFailed = Column(Integer, nullable=False)
	statusPaused = Column(Integer, nullable=False)

# create tables
base.metadata.create_all(engine)


import xlsxwriter
import logging
import json
import baangt.base.GlobalConstants as GC
from baangt.base.Timing.Timing import Timing
import sys
from baangt.base.Utils import utils
from pathlib import Path
from typing import Optional
from xlsxwriter.worksheet import (
    Worksheet, cell_number_tuple, cell_string_tuple)
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from baangt.base.DataBaseORM import DATABASE_URL, TestrunLog
from datetime import datetime
import time
from baangt import plugin_manager
import re
import csv
from dateutil.parser import parse

logger = logging.getLogger("pyC")


class ExportResults:
    def __init__(self, **kwargs):
        self.testList = []
        self.testRunInstance = kwargs.get(GC.KWARGS_TESTRUNINSTANCE)
        self.networkInfo = kwargs.get('networkInfo')
        self.testCasesEndDateTimes_1D = kwargs.get('testCasesEndDateTimes_1D')
        self.testCasesEndDateTimes_2D = kwargs.get('testCasesEndDateTimes_2D')
        self.testRunName = self.testRunInstance.testRunName
        self.dataRecords = self.testRunInstance.dataRecords

        try:
            self.exportFormat = kwargs.get(GC.KWARGS_TESTRUNATTRIBUTES).get(GC.EXPORT_FORMAT)[GC.EXPORT_FORMAT]
        except KeyError:
            self.exportFormat = GC.EXP_XLSX

        self.fileName = self.__getOutputFileName()
        logger.info("Export-Sheet for results: " + self.fileName)

        if self.exportFormat == GC.EXP_XLSX:
            self.fieldListExport = kwargs.get(GC.KWARGS_TESTRUNATTRIBUTES).get(GC.EXPORT_FORMAT)["Fieldlist"]
            self.workbook = xlsxwriter.Workbook(self.fileName)
            self.summarySheet = self.workbook.add_worksheet("Summary")
            self.worksheet = self.workbook.add_worksheet("Output")
            self.timingSheet = self.workbook.add_worksheet("Timing")
            self.cellFormatGreen = self.workbook.add_format()
            self.cellFormatGreen.set_bg_color('green')
            self.cellFormatRed = self.workbook.add_format()
            self.cellFormatRed.set_bg_color('red')
            self.cellFormatBold = self.workbook.add_format()
            self.cellFormatBold.set_bold(bold=True)
            self.summaryRow = 0
            self.__setHeaderDetailSheetExcel()
            self.makeSummaryExcel()
            self.exportResultExcel()
            self.exportTiming = ExportTiming(self.dataRecords,
                                            self.timingSheet)
            if self.networkInfo:
                self.networkSheet = self.workbook.add_worksheet("Network")
                self.exportNetWork = ExportNetWork(self.networkInfo,
                                                   self.testCasesEndDateTimes_1D,
                                                   self.testCasesEndDateTimes_2D,
                                                   self.workbook,
                                                   self.networkSheet)
            self.closeExcel()
        elif self.exportFormat == GC.EXP_CSV:
            self.export2CSV()
        self.exportToDataBase()

    # -- API support --
    def getSummary(self):
        # records status
        summary = {'Testrecords': len(self.dataRecords)}
        summary['Successful'] = len([x for x in self.dataRecords.values()
                                                   if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_SUCCESS])
        summary['Paused'] = len([x for x in self.dataRecords.values()
                                                   if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_WAITING])
        summary['Error'] = len([x for x in self.dataRecords.values()
                                                   if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR])
        # logfile
        summary['Logfile'] = logger.handlers[1].baseFilename
        # timing
        timing:Timing = self.testRunInstance.timing
        summary['Starttime'], summary['Endtime'], summary['Duration'] = timing.returnTimeSegment(GC.TIMING_TESTRUN)
        summary['Globals'] = {key: value for key, value in self.testRunInstance.globalSettings.items()}
        
        return summary
    # -- END of API support --
    
    def export2CSV(self):
        """
        Writes CSV-File of datarecords

        """
        f = open(self.fileName, 'w')
        writer = csv.DictWriter(f, self.dataRecords[0].keys())
        writer.writeheader()
        for i in range(0, len(self.dataRecords)-1):
            writer.writerow(self.dataRecords[i])
        f.close()

    def exportToDataBase(self):
        engine = create_engine(f'sqlite:///{DATABASE_URL}')

        # create a Session
        Session = sessionmaker(bind=engine)
        session = Session()

        # get timings
        timing: Timing = self.testRunInstance.timing
        start, end, duration = timing.returnTimeSegment(GC.TIMING_TESTRUN)

        # get status
        success = 0
        error = 0
        waiting = 0
        for value in self.dataRecords.values():
            if value[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_SUCCESS:
                success += 1
            elif value[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR:
                error += 1
            if value[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_WAITING:
                waiting += 1

        # get globals
        globalString = '{'
        for key, value in self.testRunInstance.globalSettings.items():
            if len(globalString) > 1:
                globalString += ', '
            globalString += f'{key}: {value}'
        globalString += '}'

        # get documents
        datafiles = self.fileName

        # create object
        log = TestrunLog(
            testrunName = self.testRunName,
            logfileName = logger.handlers[1].baseFilename,
            startTime = datetime.strptime(start, "%H:%M:%S"),
            endTime = datetime.strptime(end, "%H:%M:%S"),
            statusOk = success,
            statusFailed = error,
            statusPaused = waiting,
            globalVars = globalString,
            dataFile = datafiles,
        )
        # write to DataBase
        session.add(log)
        session.commit()

    def exportResultExcel(self, **kwargs):
        self._exportData()

    def makeSummaryExcel(self):

        self.summarySheet.write(0,0, f"Testreport for {self.testRunName}", self.cellFormatBold)
        self.summarySheet.set_column(0, last_col=0, width=15)
        # get testrunname my
        self.testList.append(self.testRunName)
        # Testrecords
        self.__writeSummaryCell("Testrecords", len(self.dataRecords), row=2, format=self.cellFormatBold)
        value = len([x for x in self.dataRecords.values()
                                                   if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_SUCCESS])
        self.testList.append(value) # Ok my
        if not value:
            value = ""
        self.__writeSummaryCell("Successful", value, format=self.cellFormatGreen)
        self.testList.append(value)  # paused my
        self.__writeSummaryCell("Paused", len([x for x in self.dataRecords.values()
                                                   if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_WAITING]))
        value = len([x for x in self.dataRecords.values()
                                               if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR])
        self.testList.append(value)  # error my
        if not value:
            value = ""
        self.__writeSummaryCell("Error", value, format=self.cellFormatRed)

        # Logfile
        self.__writeSummaryCell("Logfile", logger.handlers[1].baseFilename, row=7)
        # get logfilename for database my
        self.testList.append(logger.handlers[1].baseFilename)
        # Timing
        timing:Timing = self.testRunInstance.timing
        start, end, duration = timing.returnTimeSegment(GC.TIMING_TESTRUN)
        self.__writeSummaryCell("Starttime", start, row=9)
        # get start end during time my
        self.testList.append(start)
        self.testList.append(end)

        self.__writeSummaryCell("Endtime", end)
        self.__writeSummaryCell("Duration", duration, format=self.cellFormatBold )
        self.__writeSummaryCell("Avg. Dur", "")
        # Globals:
        self.__writeSummaryCell("Global settings for this testrun", "", format=self.cellFormatBold, row=14)
        for key, value in self.testRunInstance.globalSettings.items():
            self.__writeSummaryCell(key, str(value))
            # get global data my
            self.testList.append(str(value))
        # Testcase and Testsequence setting
        self.__writeSummaryCell("TestSequence settings follow:", "", row=16+len(self.testRunInstance.globalSettings),
                                format=self.cellFormatBold)
        lSequence = self.testRunInstance.testRunUtils.getSequenceByNumber(testRunName=self.testRunName, sequence="1")
        if lSequence:
            for key, value in lSequence[1].items():
                if isinstance(value, list) or isinstance(value, dict):
                    continue
                self.__writeSummaryCell(key, str(value))

    def __writeSummaryCell(self, lineHeader, lineText, row=None, format=None):
        if not row:
            self.summaryRow += 1
        else:
            self.summaryRow = row

        if not lineText:
            # If we have no lineText we want to apply format to the Header
            self.summarySheet.write(self.summaryRow, 0, lineHeader, format)
        else:
            self.summarySheet.write(self.summaryRow, 0, lineHeader)
            self.summarySheet.write(self.summaryRow, 1, lineText, format)

    def __getOutputFileName(self):
        if self.testRunInstance.globalSettings[GC.PATH_ROOT]:
            basePath = Path(self.testRunInstance.globalSettings[GC.PATH_ROOT])
        elif "/" not in self.testRunInstance.globalSettings[GC.DATABASE_EXPORTFILENAMEANDPATH][0:1]:
            basePath = Path(sys.modules['__main__'].__file__).parent
        else:
            basePath = ""
        l_file: Path = Path(basePath).joinpath(self.testRunInstance.globalSettings[GC.DATABASE_EXPORTFILENAMEANDPATH])
        if "~" in str(l_file.absolute()):
            l_file = l_file.expanduser()
        if not Path(l_file).is_dir():
            logger.info(f"Create directory {l_file}")
            Path(l_file).mkdir(parents=True, exist_ok=True)

        if self.exportFormat == GC.EXP_XLSX:
            lExtension = '.xlsx'
        elif self.exportFormat == GC.EXP_CSV:
            lExtension = '.csv'
        else:
            logger.critical(f"wrong export file format: {self.exportFormat}, using 'xlsx' instead")
            lExtension = '.xlsx'

        l_file = l_file.joinpath("baangt_" + self.testRunName + "_" + utils.datetime_return() + lExtension)
        logger.debug(f"Filename for export: {str(l_file)}")
        return str(l_file)

    def __setHeaderDetailSheetExcel(self):
        i = 0
        self.__extendFieldList()  # Add fields with name "RESULT_*" to output fields.
        for column in self.fieldListExport:
            self.worksheet.write(0, i, column)
            i += 1
        self.worksheet.write(0, len(self.fieldListExport), "JSON")

    def __extendFieldList(self):
        """
        Fields, that start with "RESULT_" shall always be exported.

        Other fields, that shall always be exported are also added

        @return:
        """
        lAlwaysExportFields: ["TCERRORLOG", "SCREENSHOTS"]

        try:
            for key in self.dataRecords[0].keys():
                if "RESULT_" in key.upper():
                    if not key in self.fieldListExport:
                        self.fieldListExport.append(key)
                if key.upper() in lAlwaysExportFields:
                    self.fieldListExport.append(key)

        except Exception as e:
            logger.critical(f'looks like we have no data in records: {self.dataRecords}, len of dataRecords: {len(self.dataRecords)}')

    def _exportData(self):
        for key, value in self.dataRecords.items():
            for (n, column) in enumerate(self.fieldListExport):
                self.__writeCell(key+1, n, value, column)
            # Also write everything as JSON-String into the last column
            self.worksheet.write(key+1, len(self.fieldListExport), json.dumps(value))

        # Create autofilter
        self.worksheet.autofilter(0,0,len(self.dataRecords.items()),len(self.fieldListExport)-1)

        # Make cells wide enough
        for n in range(0,len(self.fieldListExport)):
            ExcelSheetHelperFunctions.set_column_autowidth(self.worksheet, n)

    def __writeCell(self, line, cellNumber, testRecordDict, fieldName, strip=False):
        if fieldName in testRecordDict.keys() and testRecordDict[fieldName]:
            # Remove leading New-Line:
            if '\n' in testRecordDict[fieldName][0:5] or strip:
                testRecordDict[fieldName] = testRecordDict[fieldName].strip()
            # Do different stuff for Dicts and Lists:
            if isinstance(testRecordDict[fieldName], dict):
                self.worksheet.write(line, cellNumber, testRecordDict[fieldName])
            elif isinstance(testRecordDict[fieldName], list):
                self.worksheet.write(line, cellNumber,
                                     utils.listToString(testRecordDict[fieldName]))
            else:
                if fieldName == GC.TESTCASESTATUS:
                    if testRecordDict[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_SUCCESS:
                        self.worksheet.write(line, cellNumber, testRecordDict[fieldName], self.cellFormatGreen)
                    elif testRecordDict[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR:
                        self.worksheet.write(line, cellNumber, testRecordDict[fieldName], self.cellFormatRed)
                else:
                    self.worksheet.write(line, cellNumber, testRecordDict[fieldName])

    def closeExcel(self):
        self.workbook.close()
        # Next line doesn't work on MAC. Returns "not authorized"
        # subprocess.Popen([self.filename], shell=True)


class ExcelSheetHelperFunctions:
    def __init__(self):
        pass

    @staticmethod
    def set_column_autowidth(worksheet: Worksheet, column: int):
        """
        Set the width automatically on a column in the `Worksheet`.
        !!! Make sure you run this function AFTER having all cells filled in
        the worksheet!
        """
        maxwidth = ExcelSheetHelperFunctions.get_column_width(worksheet=worksheet, column=column)
        if maxwidth is None:
            return
        elif maxwidth > 45:
            maxwidth = 45
        worksheet.set_column(first_col=column, last_col=column, width=maxwidth)

    @staticmethod
    def get_column_width(worksheet: Worksheet, column: int) -> Optional[int]:
        """Get the max column width in a `Worksheet` column."""
        strings = getattr(worksheet, '_ts_all_strings', None)
        if strings is None:
            strings = worksheet._ts_all_strings = sorted(
                worksheet.str_table.string_table,
                key=worksheet.str_table.string_table.__getitem__)
        lengths = set()
        for row_id, colums_dict in worksheet.table.items():  # type: int, dict
            data = colums_dict.get(column)
            if not data:
                continue
            if type(data) is cell_string_tuple:
                iter_length = len(strings[data.string])
                if not iter_length:
                    continue
                lengths.add(iter_length)
                continue
            if type(data) is cell_number_tuple:
                iter_length = len(str(data.number))
                if not iter_length:
                    continue
                lengths.add(iter_length)
        if not lengths:
            return None
        return max(lengths)


class ExportNetWork:

    headers = ['BrowserName', 'TestCaseNum', 'Status', 'Method', 'URL', 'ContentType', 'ContentSize', 'Headers',
               'Params', 'Response', 'startDateTime', 'Duration/ms']

    def __init__(self, networkInfo: dict, testCasesEndDateTimes_1D: list,
                 testCasesEndDateTimes_2D: list, workbook: xlsxwriter.Workbook, sheet: xlsxwriter.worksheet):

        self.networkInfo = networkInfo
        self.testCasesEndDateTimes_1D = testCasesEndDateTimes_1D
        self.testCasesEndDateTimes_2D = testCasesEndDateTimes_2D
        self.workbook = workbook
        self.sheet = sheet
        header_style = self.get_header_style()
        self.write_header(style=header_style)
        self.set_column_align()
        self.write_content()
        self.set_column_width()

    def set_column_align(self):
        right_align_indexes = list()
        right_align_indexes.append(ExportNetWork.headers.index('ContentSize'))
        right_align_indexes.append(ExportNetWork.headers.index('Duration/ms'))
        right_align_style = self.get_column_style(alignment='right')
        left_align_style = self.get_column_style(alignment='left')
        [self.sheet.set_column(i, i, cell_format=right_align_style) if i in right_align_indexes else
         self.sheet.set_column(i, i, cell_format=left_align_style) for i in range(len(ExportNetWork.headers))]

    def set_column_width(self):
        [ExcelSheetHelperFunctions.set_column_autowidth(self.sheet, i) for i in range(len(ExportNetWork.headers))]

    def get_header_style(self):
        header_style = self.workbook.add_format()
        header_style.set_bg_color("#00CCFF")
        header_style.set_color("#FFFFFF")
        header_style.set_bold()
        header_style.set_border()
        return header_style

    def get_column_style(self, alignment=None):
        column_style = self.workbook.add_format()
        column_style.set_color("black")
        column_style.set_align('right') if alignment == 'right'\
            else column_style.set_align('left') if alignment == 'left' else None
        column_style.set_border()
        return column_style

    def write_header(self, style=None):
        for index, value in enumerate(ExportNetWork.headers):
            self.sheet.write(0, index, value, style)

    def _get_test_case_num(self, start_date_time, browser_name):
        d_t = parse(start_date_time)
        d_t = d_t.replace(tzinfo=None)
        if self.testCasesEndDateTimes_1D:
            for index, dt_end in enumerate(self.testCasesEndDateTimes_1D):
                if d_t < dt_end:
                    return index + 1
        elif self.testCasesEndDateTimes_2D:
            browser_num = re.findall(r"\d+\.?\d*", str(browser_name))[-1] \
                if re.findall(r"\d+\.?\d*", str(browser_name)) else 0
            dt_list_index = int(browser_num) if int(browser_num) > 0 else 0
            for i, tcAndDtEnd in enumerate(self.testCasesEndDateTimes_2D[dt_list_index]):
                if d_t < tcAndDtEnd[1]:
                    return tcAndDtEnd[0] + 1
        return 'unknown'

    def write_content(self):
        if not self.networkInfo:
            return


        partition_index = 0

        for info in self.networkInfo:
            for index, entry in enumerate(info['log']['entries']):
                browser_name = entry['pageref']
                status = entry['response']['status']
                method = entry['request']['method']
                url = entry['request']['url']
                content_type = entry['response']['content']['mimeType']
                content_size = entry['response']['content']['size']
                headers = entry['response']['headers']
                params = entry['request']['queryString']
                response = entry['response']['content']['text'] if 'text' in entry['response']['content'] else ''
                start_date_time = entry['startedDateTime']
                duration = entry['time']
                test_case_num = self._get_test_case_num(start_date_time, browser_name)

                data_list = [browser_name, test_case_num, status, method, url, content_type, content_size,
                             headers, params, response, start_date_time, duration]

                [self.sheet.write(index + partition_index + 1, i, str(data_list[i]) or 'null')
                 for i in range(len(data_list))]

            partition_index += len(info['log']['entries'])



class ExportTiming:
    def __init__(self, testdataRecords:dict, sheet:xlsxwriter.worksheet):
        self.testdataRecords = testdataRecords
        self.sheet:xlsxwriter.worksheet = sheet

        self.sections = {}

        self.findAllTimingSections()
        self.writeHeader()
        self.writeLines()

        # Autowidth
        for n in range(0,len(self.sections)+1):
            ExcelSheetHelperFunctions.set_column_autowidth(self.sheet, n)

    def writeHeader(self):
        self.wc(0,0,"Testcase#")
        for index, key in enumerate(self.sections.keys(), start=1):
            self.wc(0, index, key)

    def writeLines(self):
        for tcNumber, (key, line) in enumerate(self.testdataRecords.items(),start=1):
            self.wc(tcNumber, 0, tcNumber)
            lSections = self.interpretTimeLog(line[GC.TIMELOG])
            for section, timingValue in lSections.items():
                # find, in which column this section should be written:
                for column, key in enumerate(self.sections.keys(),1):
                    if key == section:
                        self.wc(tcNumber, column,
                                timingValue[GC.TIMING_DURATION])
                        continue

    @staticmethod
    def shortenTimingValue(timingValue):
        # TimingValue is seconds in Float. 2 decimals is enough:
        timingValue = int(float(timingValue) * 100)
        return timingValue/100

    def writeCell(self, row, col, content, format=None):
        self.sheet.write(row, col, content, format)

    wc = writeCell

    def findAllTimingSections(self):
        """
        We try to have an ordered list of Timing Sequences. As each Testcase might have different sections we'll have
        to make guesses

        @return:
        """
        lSections = {}
        for key, line in self.testdataRecords.items():
            lTiming:dict = ExportTiming.interpretTimeLog(line[GC.TIMELOG])
            for key in lTiming.keys():
                if lSections.get(key):
                    continue
                else:
                    lSections[key] = None

        self.sections = lSections

    @staticmethod
    def interpretTimeLog(lTimeLog):
        """Example Time Log:
        Complete Testrun: Start: 1579553837.241974 - no end recorded
        TestCaseSequenceMaster: Start: 1579553837.243414 - no end recorded
        CustTestCaseMaster: Start: 1579553838.97329 - no end recorded
        Browser Start: , since last call: 2.3161418437957764
        Empfehlungen: , since last call: 6.440968036651611, ZIDs:[175aeac023237a73], TS:2020-01-20 21:57:46.525577
        Annahme_RABAZ: , since last call: 2.002716064453125e-05, ZIDs:[6be7d0a44e59acf6], TS:2020-01-20 21:58:37.203583
        Antrag drucken: , since last call: 9.075241088867188, ZIDs:[6be7d0a44e59acf6, b27c3875ddcbb4fa], TS:2020-01-20 21:58:38.040137
        Warten auf Senden an Bestand Button: , since last call: 1.3927149772644043
        Senden an Bestand: , since last call: 9.60469913482666, ZIDs:[66b12fa4869cf8a0, ad1f3d47c4694e26], TS:2020-01-20 21:58:49.472288

        where the first part before ":" is the section, "since last call:" is the duration, TS: is the timestamp

        Update 29.3.2020: Format changed to "since last call: 00:xx:xx,", rest looks identical.
        """
        lExport = {}
        lLines = lTimeLog.split("\n")
        for line in lLines:
            parts = line.split(",")
            if len(parts) < 2:
                continue
            if "Start:" in line:
                # Format <sequence>: <Start>: <time.loctime>
                continue
            else:
                lSection = parts[0].replace(":","").strip()
                lDuration = parts[1].split("since last call: ")[1]
                lExport[lSection] = {GC.TIMING_DURATION: lDuration}
        return lExport




import baangt.base.GlobalConstants as GC
from baangt.base.Timing.Timing import Timing
from baangt.base.BrowserHandling.BrowserHandling import BrowserDriver
from baangt.base.ApiHandling import ApiHandling
import sys
from pkg_resources import parse_version
import logging
from baangt.TestSteps.Exceptions import baangtTestStepException
from baangt.TestSteps.AddressCreation import AddressCreate

logger = logging.getLogger("pyC")


class TestStepMaster:
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.testRunInstance = kwargs.get(GC.KWARGS_TESTRUNINSTANCE)
        self.testcaseDataDict = kwargs.get(GC.KWARGS_DATA)
        self.timing: Timing = kwargs.get(GC.KWARGS_TIMING)
        self.timingName = self.timing.takeTime(self.__class__.__name__, forceNew=True)
        self.browserSession: BrowserDriver = kwargs.get(GC.KWARGS_BROWSER)
        self.apiSession: ApiHandling = kwargs.get(GC.KWARGS_API_SESSION)
        self.testCaseStatus = None
        self.testStepNumber = kwargs.get(GC.STRUCTURE_TESTSTEP)    # Set in TestRun by TestCaseMaster
        self.testRunUtil = self.testRunInstance.testRunUtils
        # check, if this TestStep has additional Parameters and if so, execute
        lSequence = self.testRunUtil.getSequenceByNumber(testRunName=self.testRunInstance.testRunName,
                                                         sequence=kwargs.get(GC.STRUCTURE_TESTCASESEQUENCE))
        lTestCase = self.testRunUtil.getTestCaseByNumber(lSequence, kwargs.get(GC.STRUCTURE_TESTCASE))
        lTestStep = self.testRunUtil.getTestStepByNumber(lTestCase, kwargs.get(GC.STRUCTURE_TESTSTEP))
        self.globalRelease = self.testRunInstance.globalSettings.get("Release", "")
        self.ifActive = False
        self.ifIsTrue = True

        if not isinstance(lTestStep, str):
            # This TestStepMaster-Instance should actually do something - activitites are described
            # in the TestExecutionSteps
            self.executeDirect(lTestStep[1][GC.STRUCTURE_TESTSTEPEXECUTION])

        self.teardown()

    def executeDirect(self, executionCommands):
        """
        This will execute single Operations directly
        """
        for key, command in executionCommands.items():
            # when we have an IF-condition and it's condition was not TRUE, then skip whatever comes here until we
            # reach Endif
            if not self.ifIsTrue and command["Activity"] != "ENDIF":
                continue
            lActivity = command["Activity"].upper()
            if lActivity == "COMMENT":
                continue     # Comment's are ignored

            lLocatorType = command["LocatorType"].upper()
            lLocator = self.replaceVariables(command["Locator"])

            if lLocator and not lLocatorType:   # If locatorType is empty, default it to XPATH
                lLocatorType = 'XPATH'

            xpath, css, id = self.__setLocator(lLocatorType, lLocator)

            lValue = str(command["Value"])
            lValue2 = str(command["Value2"])
            lComparison = command["Comparison"]
            lOptional = TestStepMaster._sanitizeXField(command["Optional"])

            # check release line
            lRelease = command["Release"]

            lTimeout = TestStepMaster.__setTimeout(command["Timeout"])

            # Replace variables from data file
            if len(lValue) > 0:
                lValue = self.replaceVariables(lValue)
            if len(lValue2) > 0:
                lValue2 = self.replaceVariables(lValue2)

            if not TestStepMaster.ifQualifyForExecution(self.globalRelease, lRelease):
                logger.debug(f"we skipped this line due to {lRelease} disqualifies according to {self.globalRelease} ")
                continue  # We ignored the steps as it doesn't qualify
            if lActivity == "GOTOURL":
                self.browserSession.goToUrl(lValue)
            elif lActivity == "SETTEXT":
                self.browserSession.findByAndSetText(xpath=xpath, css=css, id=id, value=lValue, timeout=lTimeout)
            elif lActivity == 'HANDLEIFRAME':
                self.browserSession.handleIframe(lLocator)
            elif lActivity == "CLICK":
                self.browserSession.findByAndClick(xpath=xpath, css=css, id=id, timeout=lTimeout)
            elif lActivity == "PAUSE":
                self.browserSession.sleep(sleepTimeinSeconds=lValue)
            elif lActivity == "IF":
                if self.ifActive:
                    raise BaseException("No nested IFs at this point, sorry...")
                self.ifActive = True
                # Originally we had only Comparisons. Now we also want to check for existance of Field
                if not lValue and lLocatorType and lLocator:
                    lValue = self.browserSession.findBy(xpath=xpath, css=css, id=id, optional=lOptional,
                                                        timeout=lTimeout)

                self.__doComparisons(lComparison=lComparison, value1=lValue, value2=lValue2)
            elif lActivity == "ENDIF":
                if not self.ifActive:
                    raise BaseException("ENDIF without IF")
                self.ifActive = False
                self.ifIsTrue = True
            elif lActivity == 'GOBACK':
                self.browserSession.goBack()
            elif lActivity == 'APIURL':
                self.apiSession.setBaseURL(lValue)
            elif lActivity == 'ENDPOINT':
                self.apiSession.setEndPoint(lValue)
            elif lActivity == 'POST':
                self.apiSession.postURL(content=lValue)
            elif lActivity == 'GET':
                self.apiSession.getURL()
            elif lActivity == 'HEADER':
                self.apiSession.setHeaders(setHeaderData=lValue)
            elif lActivity == 'SAVE':
                self.doSaveData(lValue, lValue2)
            elif lActivity == 'SUBMIT':
                self.browserSession.submit()
            elif lActivity == "ADDRESS_CREATE":
                # Create Address with option lValue and lValue2
                AddressCreate(lValue,lValue2)
                # Update testcaseDataDict with addressDict returned from
                AddressCreate.returnAddress()
                self.testcaseDataDict.update(AddressCreate.returnAddress())
            elif lActivity == 'ASSERT':
                value_found = self.browserSession.findByAndWaitForValue(xpath=xpath, css=css, id=id, optional=lOptional,
                                                        timeout=lTimeout)
                value_expected = lValue
                if value_expected != value_found:
                    raise baangtTestStepException(f"Expected Value: {value_expected}, Value found :{value_found} ")
            elif lActivity == 'IBAN':
                # Create Random IBAN. Value1 = Input-Parameter for IBAN-Function. Value2=Fieldname
                self.__getIBAN(lValue, lValue2)
            elif lActivity == 'PDFCOMPARE':
                lFiles = self.browserSession.findNewFiles()
                # fixme: Implement the API-Call here
            elif lActivity == 'CHECKLINKS':
                self.checkLinks()
            else:
                raise BaseException(f"Unknown command in TestStep {lActivity}")

    def __getIBAN(self, lValue, lValue2):
        from baangt.base.IBAN import IBAN
        if not lValue2:
            logger.critical("IBAN-Method called without destination field name in column 'Value 2'")
            return

        lIBAN = IBAN()
        self.testcaseDataDict[lValue2] = lIBAN.getRandomIBAN()
        pass

    def checkLinks(self):
        """
        Will check all links on the current webpage

        Result will be written into "CheckedLinks" in TestDataDict

        If theres a returncode >= 400 in the list, we'll mark the testcase as failed
        """
        if self.testcaseDataDict.get("CheckedLinks"):
            self.testcaseDataDict["Result_CheckedLinks"] += self.browserSession.checkLinks()
        else:
            self.testcaseDataDict["Result_CheckedLinks"] = self.browserSession.checkLinks()

        for entry in self.testcaseDataDict["Result_CheckedLinks"]:
            if entry[0] > 400:
                self.testcaseDataDict[GC.TESTCASESTATUS] = GC.TESTCASESTATUS_ERROR
                self.testcaseDataDict[GC.TESTCASEERRORLOG] = self.testcaseDataDict.get(GC.TESTCASEERRORLOG,"")\
                                                             + "URL-Checker error"
                break

    @staticmethod
    def _sanitizeXField(inField):
        """
        When "X" or "True" is sent, then use this
        @param inField:
        @return:
        """
        lXField = True

        if not inField:
            lXField = False

        if inField.upper() == 'FALSE':
            lXField = False

        return lXField

    @staticmethod
    def ifQualifyForExecution(version_global, version_line):
        """ This function will test version_global and version_line
            @return True or False
        """
        if not version_global.strip():
            # No value is defined in Release, return True
            return True
        if not version_line.strip():
            # we skipped this line
            return True

        # split the version line
        if not len(version_line.strip().split(" ")) == 2:
            logger.debug(f"Invalid release format {version_line} ")
            return True
        comp_operator, version = version_line.strip().split(" ")
        if comp_operator == "<":
            return parse_version(version_global) < parse_version(version)
        elif comp_operator == ">":
            return parse_version(version_global) > parse_version(version)
        elif comp_operator == ">=":
            return parse_version(version_global) >= parse_version(version)
        elif comp_operator == "<=":
            return parse_version(version_global) <= parse_version(version)
        elif comp_operator == "=" or comp_operator == "==":
            return parse_version(version_global) == parse_version(version)
        else:
            logger.debug(f"Global version {version_global}, line version {version_line} ")
            return False

    def doSaveData(self, toField, valueForField):
        self.testcaseDataDict[toField] = valueForField

    @staticmethod
    def __setLocator(lLocatorType, lLocator):
        """

        @param lLocatorType: XPATH, CSS, ID, etc.
        @param lLocator: Value of the locator
        @return:
        """
        xpath = None
        css = None
        lId = None

        if lLocatorType:
            if lLocatorType == 'XPATH':
                xpath = lLocator
            elif lLocatorType == 'CSS':
                css = lLocator
            elif lLocatorType == 'ID':
                lId = lLocator

        return xpath, css, lId

    @staticmethod
    def __setTimeout(lTimeout):
        if lTimeout:
            lTimeout = float(lTimeout)
        else:
            lTimeout = 20
        return lTimeout

    @staticmethod
    def anyting2Boolean(valueIn):
        if isinstance(valueIn, bool):
            return valueIn

        if isinstance(valueIn, int):
            return bool(valueIn)

        if isinstance(valueIn, str):
            if valueIn.lower() in ("yes", "true", "1", "ok"):
                return True
            else:
                return False

        raise TypeError(f"Anything2Boolean had a wrong value: {valueIn}. Don't know how to convert that to boolean")

    def __doComparisons(self, lComparison, value1, value2):
        if isinstance(value1, bool) or isinstance(value2, bool):
            value1 = TestStepMaster.anyting2Boolean(value1)
            value2 = TestStepMaster.anyting2Boolean(value2)

        if lComparison == "=":
            if value1 == value2:
                self.ifIsTrue = True
            else:
                self.ifIsTrue = False
        elif lComparison == ">":
            if value1 > value2:
                self.ifIsTrue = True
            else:
                self.ifIsTrue = False
        elif lComparison == "<":
            if value1 < value2:
                self.ifIsTrue = True
            else:
                self.ifIsTrue = False
        else:
            raise BaseException(f"Comparison Operator not supported/unknown {lComparison}")

    def execute(self):
        """Method is overwritten in all children/subclasses"""
        pass

    def teardown(self):
        if self.testCaseStatus:
            if not self.testcaseDataDict[GC.TESTCASESTATUS]:
                self.testcaseDataDict[GC.TESTCASESTATUS] = self.testCaseStatus
            elif self.testCaseStatus == GC.TESTCASESTATUS_SUCCESS:
                # We'll not overwrite a potential Error Status of the testcase
                pass
            elif self.testCaseStatus == GC.TESTCASESTATUS_ERROR:
                self.testcaseDataDict = GC.TESTCASESTATUS_ERROR
            else:
                sys.exit("No idea, what happened here. Unknown condition appeared")

        self.timing.takeTime(self.timingName)

    def replaceVariables(self, expression):
        """
        The syntax for variables is currently $(<column_name_from_data_file>). Multiple variables can be assigned
        in one cell, for instance perfectly fine: "http://$(BASEURL)/$(ENDPOINT)"

        @param expression: the current cell, either as fixed value, e.g. "Franzi" or with a varible $(DATE)
        @return: the replaced value, e.g. if expression was $(DATE) and the value in column "DATE" of data-file was
            "01.01.2020" then return will be "01.01.2020"
        """
        if "$(" not in expression:
            return expression

        while "$(" in expression:
            if expression[0:2] == "$(":
                left_part = ""
            else:
                left_part = expression.split("$(")[0]

            center = expression[len(left_part)+2:]
            center = center.split(")")[0]

            right_part = expression[len(left_part)+len(center)+3:]

            if "." not in center:
                # Replace the variable with the value from data structure
                center = self.testcaseDataDict.get(center)
            else:
                # This is a reference to a DICT with ".": for instance APIHandling.AnswerContent("<bla>")
                dictVariable = center.split(".")[0]
                dictValue = center.split(".")[1]

                if dictVariable == 'ANSWER_CONTENT':
                    center = self.apiSession.session[1].answerJSON.get(dictValue, "Empty")
                else:
                    raise BaseException(f"Missing code to replace value for: {center}")

            if not center:
                raise BaseException(f"Variable not found: {center}")

            expression = "".join([left_part, center, right_part])
        return expression



import PySimpleGUI as sg
import os
import xlsxwriter
from pathlib import Path
import pyperclip
import logging
import re
from baangt.base.ExportResults.ExportResults import ExcelSheetHelperFunctions

# re for to extract 'text' from  //input[@type='text']
extract_var = r'.*\'(.*)\''

logger = logging.getLogger("pyC")


class ImportKatalonRecorder:
    def __init__(self, directory):
        self.window = None
        self.fileNameExport = None
        self.directory = directory
        self.clipboardText = ""
        self.outputText = ""
        self.outputData = {}

        self.clipboardText = """
        
Please
        
export the recording 
        
in 
        
Katalon Recorder (Firefox or Chrome)
        
and
        
click "Import Clipboard" button below.
   
        
Alternatively copy + paste manually
into this side of the window.
"""
        self.outputFormatted = []
        self.startWindow()

    def startWindow(self):
        self.getLayout()
        sg.theme("TanBlue")

        self.window = sg.Window("Baangt interactive Starter", layout=self.getLayout())

        lWindow = self.window
        lWindow.finalize()

        while True:
            lEvent, lValues = lWindow.read()
            if lEvent == "Exit":
                break

            if lEvent == 'Save':
                self.fileNameExport = sg.popup_get_text("Name of new Testcase:")
                if self.fileNameExport:
                    self.saveTestCase()
                    break

            if lEvent == "TextIn":  # Textinput in TextIn
                self.__importTranslateAndUpdate(lWindow=lWindow)

            if lEvent == 'Import from Clipboard':
                self.clipboardText = pyperclip.paste()
                self.window["TextIn"].update(value=self.clipboardText)
                self.__importTranslateAndUpdate(lWindow=lWindow)

        lWindow.close()

    def __importTranslateAndUpdate(self, lWindow):
        self.importClipboard()
        lWindow['TextOut'].update(value=self.outputText)

    def saveTestCase(self):
        if "XLSX" not in self.fileNameExport.upper():
            self.fileNameExport = self.fileNameExport + ".xlsx"
        lExcel = xlsxwriter.Workbook(str(Path(self.directory).joinpath(self.fileNameExport)))
        lWorksheetDefinition = lExcel.add_worksheet("TestStepExecution")
        lWorksheetData = lExcel.add_worksheet("data")
        self.saveTestCaseHeader(worksheet=lWorksheetDefinition)
        self.saveTestCaseExecution(worksheet=lWorksheetDefinition)
        self.saveTestData(lWorksheetData)
        lExcel.close()

    def saveTestData(self, worksheet):
        # fixed: Translated the values in Testcase with VariableNames and then write those to data.
        for index, data in enumerate(sorted(self.outputData.items())):
            # remove $( and ) from header
            header = re.search(r'.*\((.*)\)', data[0]).groups()[0]
            self.writeCell(worksheet, 0, index, header)
            self.writeCell(worksheet, 1, index, data[1])

        for i in range(len(self.outputData)):
            ExcelSheetHelperFunctions.set_column_autowidth(worksheet, i)

    def saveTestCaseHeader(self, worksheet):
        self.writeCell(worksheet, 0, 0, "Activity")
        self.writeCell(worksheet, 0, 1, "LocatorType")
        self.writeCell(worksheet, 0, 2, "Locator")
        self.writeCell(worksheet, 0, 3, "Value")
        self.writeCell(worksheet, 0, 4, "Comparison")
        self.writeCell(worksheet, 0, 5, "Value2")
        self.writeCell(worksheet, 0, 6, "Timeout")
        self.writeCell(worksheet, 0, 7, "Optional")
        self.writeCell(worksheet, 0, 8, "Release")
        # fixme: Write comments in the header line

    def saveTestCaseExecution(self, worksheet):
        for row, line in enumerate(self.outputFormatted, start=1):
            if not line:
                continue
            for col, (key, value) in enumerate(line.items()):
                self.writeCell(worksheet, row, col, value)
        for i in range(len(self.outputFormatted)):
            ExcelSheetHelperFunctions.set_column_autowidth(worksheet, i)

    def writeCell(self, sheet, cellRow, cellCol, value, format=None):
        sheet.write(cellRow, cellCol, value)

    def getLayout(self):
        lLayout = []
        lLayout.append([sg.Multiline(default_text=self.clipboardText, size=(50, 30), key='TextIn', change_submits=True),
                        sg.Multiline(default_text=self.outputText, size=(50, 30), key='TextOut')])
        lLayout.append([sg.Text("Select Directory, Testrun and Global settings to use:")])
        lLayout.append([sg.Text("Directory", size=(15, 1)),
                        sg.In(key="-directory-", size=(30, 1), enable_events=True, default_text=self.directory),
                        sg.FolderBrowse(initial_folder=os.getcwd(), enable_events=True)])

        lLayout.append([sg.Button('Save'), sg.Button('Exit'), sg.Button("Import from Clipboard")])

        return lLayout

    def importClipboard(self):
        lineOut = []

        for line in self.clipboardText.split("\n"):
            if len(line.strip()) > 2:
                if "|" not in line:
                    continue

                lineOut.append(self.doTranslate(line))
        # Further process into $('variable') and value format
        lineOut, self.outputData = ImportKatalonRecorder.splitVariable(lineOut)

        self.outputText = ""
        self.outputFormatted = lineOut
        for line in lineOut:
            if not line:
                continue
            lStr = ""
            for key, value in line.items():
                lStr = lStr + f"{key}: {value}, "
            lStr = lStr + "\n"
            self.outputText += lStr

    def doTranslate(self, lineIn):
        value = ""
        if "|" not in lineIn:
            return None
        lParts = lineIn.split("|")
        command = lParts[0].strip().lower()
        locator = lParts[1].strip()
        if command == "type" or command == "sendKeys":
            value = lParts[2].strip()

        if command == 'click':
            return ImportKatalonRecorder.doTranslateClick(locator)
        elif command == 'type':
            return ImportKatalonRecorder.doTranslateType(locator, value)
        elif command == 'sendkeys':
            return ImportKatalonRecorder.doTranslateType(locator, value)
        elif command == 'open':
            return ImportKatalonRecorder.__fillDict("GOTOURL", "", locator)
        elif command == 'gobackandwait' or command == 'goback':
            return ImportKatalonRecorder.doTranslategoBackAndWait()
        elif command == 'select':
            return ImportKatalonRecorder.doTranslateSelect(locator)
        elif command == 'submit':
            return ImportKatalonRecorder.doTranslateSubmit(locator)
        else:
            logger.exception(f"Translation for command not implemented: {command}")
            return None

    @staticmethod
    def doTranslateType(locator, value):
        return ImportKatalonRecorder.__fillDict("SETTEXT", locator, value)

    @staticmethod
    def doTranslateClick(locator):
        return ImportKatalonRecorder.__fillDict("click", locator)

    @staticmethod
    def doTranslategoBackAndWait():
        return ImportKatalonRecorder.__fillDict("goBack", "")

    @staticmethod
    def doTranslateSelect(locator):
        return ImportKatalonRecorder.__fillDict("select", locator)

    @staticmethod
    def doTranslateSubmit(locator):
        return ImportKatalonRecorder.__fillDict("submit", locator)

    @staticmethod
    def __fillDict(activity, locator, value="", locatorType='xpath'):
        if activity == 'select':
            activity = 'click'
            return {"Activity": activity,
                    "LocatorType": locatorType,
                    "Locator": ImportKatalonRecorder.doTranslateLocator(locator, "select"),
                    "Value": str(value)}

        return {"Activity": activity,
                "LocatorType": locatorType,
                "Locator": ImportKatalonRecorder.doTranslateLocator(locator),
                "Value": str(value)
                }

    @staticmethod
    def doTranslateLocator(locator, specialInstructions=None):

        if "xpath=" in locator:
            # he comes with xpath=(//button[@type='button'])[5]
            locator = locator.replace("xpath=", "")
        if locator[0:3].upper() == "ID=":
            # This is an ID. Translate the ID to proper xpath-Syntax
            if specialInstructions == "select":
                preLocator = "//select[@id='"
            else:
                preLocator = "//*[@id='"
            postLocator = "']"
            locator = locator[3:]
            locator = preLocator + locator + postLocator
        if locator[0:5].upper() == 'LINK=':
            # //a[text()='text_i_want_to_find']/@href --> Didn't work. Return Text and Selenium can't handle that
            # //a[contains(., 'Button')]
            preLocator = "//a[contains(., '"
            postLocator = "')]"
            locator = locator[5:]
            locator = preLocator + locator + postLocator
        return locator

    @staticmethod
    def splitVariable(lines):
        # make outputData empty if already there
        """ This function will process each line and format the
            'Value' column in format $(variabe).

        @output: list(lines), list(outputData)
        """

        outputData = {}
        # loop through line and process if 'Value' column there
        for line in lines:
            # Check for GOTOURL activity
            # skip if lines == NONE
            if not line:
                continue
            if line['Activity'] == "GOTOURL":
                # we should make variable for url
                line['Value'] = ImportKatalonRecorder.prepareKeyValue(outputData, 'url', line['Value'])

            elif line['Value'].strip():
                # we will make it variable
                result = re.search(extract_var, line['Locator'])
                if result:
                    # create variable format like $(value)
                    result = result.groups()[0]

                    line['Value'] = ImportKatalonRecorder.prepareKeyValue(outputData, result, line['Value'])
        return lines, outputData

    @staticmethod
    def prepareKeyValue(outputData, key, value):
        """ This function will append key and value in 
            outputData
            if variable key exist, it will rename by
                    key1, key2, ...
        """
        for i in range(1, 100):
            var = "$(" + key + str(i) + ")"
            if not outputData.get(var, None):
                outputData[var] = value
                return var
            else:
                continue

    def exportResult(self):
        pass



from baangt.base import GlobalConstants as GC

class AddressCreate:

    returnDict = {
        GC.ADDRESS_COUNTRYCODE: "",
        GC.ADDRESS_POSTLCODE: "",
        GC.ADDRESS_CITYNAME: "",
        GC.ADDRESS_STREETNAME: "",
        GC.ADDRESS_HOUSENUMBER: "",
        GC.ADDRESS_ADDITION1: "",
        GC.ADDRESS_ADDITION2: ""
    }

    def __init__(self, addressFilterCriteria:dict):
        self.filterCriteria = addressFilterCriteria
        pass

    def returnAddress(self):
        # fixme: Implement the functionality.

        result = self.returnDict
        if self.filterCriteria.get(GC.ADDRESS_COUNTRYCODE) == 'AT' or not self.filterCriteria:
            result[GC.ADDRESS_COUNTRYCODE] = "AT"
            result[GC.ADDRESS_CITYNAME] = "Wien"
            result[GC.ADDRESS_POSTLCODE] = "1020"
            result[GC.ADDRESS_STREETNAME] = "Castellezgasse"
            result[GC.ADDRESS_HOUSENUMBER] = "32"
            result[GC.ADDRESS_ADDITION1] = "5"
        elif self.filterCriteria.get(GC.ADDRESS_COUNTRYCODE) == 'CY':
            result[GC.ADDRESS_COUNTRYCODE] = "CY"
            result[GC.ADDRESS_CITYNAME] = "Larnaca"
            result[GC.ADDRESS_POSTLCODE] = "6020"
            result[GC.ADDRESS_STREETNAME] = "Pavlou Valsamaki"
            result[GC.ADDRESS_HOUSENUMBER] = "1"
            result[GC.ADDRESS_ADDITION1] = "Apt 202"

        return self.returnDict




import baangt

from baangt.base.TestRun.TestRun import TestRun


class TestRunHookImpl(object):

    @baangt.hook_impl
    def testRun_init(self, testRunName, globalSettingsFileNameAndPath):
        return TestRun(testRunName=testRunName, globalSettingsFileNameAndPath=globalSettingsFileNameAndPath)

    @baangt.hook_impl
    def testRun_tearDown(self, testRunObject):
        return testRunObject.tearDown()

    @baangt.hook_impl
    def testRun_getSuccessAndError(self, testRunObject):
        return testRunObject.testRun_getSuccessAndError()

    @baangt.hook_impl
    def testRun_getAllTestRunAttributes(self, testRunObject):
        return testRunObject.getAllTestRunAttributes()

    @baangt.hook_impl
    def testRun_getBrowser(self, testRunObject, browserInstance=1, browserName=None, browserAttributes=None):
        return testRunObject.getBrowser(browserInstance, browserName, browserAttributes)

    @baangt.hook_impl
    def testRun_getAPI(self, testRunObject):
        return testRunObject._getAPI()

    @baangt.hook_impl
    def testRun_setResult(self, testRunObject, recordNumber, dataRecordResult):
        return testRunObject.setResult(recordNumber, dataRecordResult)

    @baangt.hook_impl
    def testRun_executeTestRun(self, testRunObject):
        return testRunObject.executeTestRun()

    @baangt.hook_impl
    def testRun_executeDictSequenceOfClasses(self, testRunObject, dictSequenceOfClasses, counterName, **kwargs):
        return testRunObject.executeDictSequenceOfClasses(dictSequenceOfClasses, counterName, **kwargs)

    @baangt.hook_impl
    def testRun_loadJSONGlobals(self, testRunObject):
        return testRunObject._loadJSONGlobals()




